import sys
import json


def render(node, strict=False):
    """Recipe to render a given FST node.

    The FST is composed of branch nodes which are either lists or dicts
    and of leaf nodes which are strings. Branch nodes can have other
    list, dict or leaf nodes as childs.

    To render a string, simply output it. To render a list, render each
    of its elements in order. To render a dict, you must follow the
    node's entry in the nodes_rendering_order dictionary and its
    dependents constraints.

    This function hides all this algorithmic complexity by returning
    a structured rendering recipe, whatever the type of node. But even
    better, you should subclass the RenderWalker which simplifies
    drastically working with the rendered FST.

    The recipe is a list of steps, each step correspond to a child and is actually a 3-uple composed of the following fields:

    - `key_type` is a string determining the type of the child in the second field (`item`) of the tuple. It can be one of:

      - 'constant': the child is a string
      - 'node': the child is a dict
      - 'key': the child is an element of a dict
      - 'list': the child is a list
      - 'formatting': the child is a list specialized in formatting

    - `item` is the child itself: either a string, a dict or a list.
    - `render_key` gives the key used to access this child from the parent node. It's a string if the node is a dict or a number if its a list.

    Please note that "bool" `key_types` are never rendered, that's why
    they are not shown here.
    """
    if isinstance(node, list):
        return render_list(node)

    elif isinstance(node, dict):
        return render_node(node, strict=strict)

    else:
        raise NotImplementedError("You tried to render a %s. Only list and dicts can be rendered." % node.__class__.__name__)


def render_list(node):
    for pos, child in enumerate(node):
        yield ('node', child, pos)


def render_node(node, strict=False):
    if node["type"] not in nodes_rendering_order:
        raise Exception("There are no defined rules for rendering a node of type '%s', has it been defined in render.py?" % node["type"])

    for key_type, render_key, dependent in nodes_rendering_order[node['type']]:
        if not dependent:
            continue
        elif key_type == "bool":
            raise NotImplementedError("Bool keys are only used for dependency, they cannot be rendered. Please set the \"%s\"'s dependent key in \"%s\" node to False" % ((key_type, render_key, dependent), node['type']))
        elif isinstance(dependent, str) and not node.get(dependent):
            continue
        elif isinstance(dependent, list) and not all([node.get(x) for x in dependent]):
            continue

        if strict:
            try:
                if key_type == "key":
                    assert isinstance(node[render_key], (dict, type(None))), "Key '%s' is expected to have type of 'key' (dict/None) but has type of '%s' instead" % (render_key, type(node[render_key]))
                elif key_type == "string":
                    assert isinstance(node[render_key], str), "Key '%s' is expected to have type of 'string' but has type of '%s' instead" % (render_key, type(node[render_key]))
                elif key_type in ("list", "formatting"):
                    assert isinstance(node[render_key], list), "Key '%s' is expected to have type of 'list' but has type of '%s' instead" % (render_key, type(node[render_key]))
                elif key_type == "constant":
                    pass
                else:
                    raise Exception("Invalid key_type '%s', should be one of those: key, string, list, formatting" % key_type)

                if dependent is True:
                    pass
                elif isinstance(dependent, str):
                    assert dependent in node
                elif isinstance(dependent, list):
                    assert all([x in node for x in dependent])
            except AssertionError as e:
                sys.stdout.write("Where node.type == '%s', render_key == '%s' and node ==\n%s\n" % (node["type"], render_key, json.dumps(node, indent=4, sort_keys=True)))
                raise e

        if key_type in ['key', 'string', 'list', 'formatting']:
            yield (key_type, node[render_key], render_key)
        elif key_type in ['constant', 'string']:
            yield (key_type, render_key, render_key)
        else:
            raise NotImplementedError("Unknown key type \"%s\" in \"%s\" node" % (key_type, node['type']))


node_types = set(['node', 'list', 'key', 'formatting', 'constant', 'bool', 'string'])


def node_keys(node):
    return [key for (_, key, _) in nodes_rendering_order[node['type']]]


def child_by_key(node, key):
    if isinstance(node, list):
        return node[key]

    if key in node:
        return node[key]

    if key in node_keys(node):
        return key

    raise AttributeError("Cannot access key \"%s\" in node \"%s\"" % (key, node))


# for a surprising exception, we won't honnor pep8 here because it really increase lisibility
nodes_rendering_order = {
    "int":                       [("string", "value", True)],  # noqa
    "long":                      [("string", "value", True)],  # noqa
    "name":                      [("string", "value", True)],  # noqa
    "hexa":                      [("string", "value", True)],  # noqa
    "octa":                      [("string", "value", True)],  # noqa
    "float":                     [("string", "value", True)],  # noqa
    "space":                     [("string", "value", True)],  # noqa
    "binary":                    [("string", "value", True)],  # noqa
    "complex":                   [("string", "value", True)],  # noqa
    "float_exponant":            [("string", "value", True)],  # noqa
    "left_parenthesis":          [("string", "value", True)],  # noqa
    "right_parenthesis":         [("string", "value", True)],  # noqa
    "float_exponant_complex":    [("string", "value", True)],  # noqa

    "break":                     [("string", "type", True)],  # noqa
    "continue":                  [("string", "type", True)],  # noqa
    "pass":                      [("string", "type", True)],  # noqa

    "dotted_name":               [("list", "value", True)],  # noqa
    "ifelseblock":               [("list", "value", True)],  # noqa
    "atomtrailers":              [("list", "value", True)],  # noqa
    "string_chain":              [("list", "value", True)],  # noqa

    "endl": [
        ("formatting", "formatting",        True),  # noqa
        ("string",     "value",             True),  # noqa
        ("string",     "indent",            True),  # noqa
    ],

    "star": [
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],

    "star_expression": [
        ("constant",   "*",                 True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("key",        "value",             True),  # noqa
    ],

    "string": [
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],
    "raw_string": [
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],
    "binary_string": [
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],
    "unicode_string": [
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],
    "binary_raw_string": [
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],
    "interpolated_string": [
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],
    "interpolated_raw_string": [
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],
    "unicode_raw_string": [
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],

    # FIXME ugly, comment can end up in formatting of another
    # node or being standalone, this is bad
    "comment": [
        ("formatting", "formatting", "formatting"),  # noqa
        ("string",     "value",             True),  # noqa
    ],

    "ternary_operator": [
        ("key",        "first",             True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "if",                True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "value",             True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   "else",              True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
        ("key",        "second",            True),  # noqa
    ],

    "ellipsis": [
        ("constant",   ".",                 True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ".",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   ".",                 True),  # noqa
    ],

    "dot": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ".",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],

    "semicolon": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ";",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],

    "comma": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ",",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
    ],

    "call": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "(",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("list",       "value",             True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   ")",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "decorator": [
        ("constant",   "@",                 True),  # noqa
        ("key",        "value",             True),  # noqa
        ("key",        "call",              "call"),  # noqa
    ],

    "class": [
        ("list",       "decorators",        True),  # noqa
        ("constant",   "class",             True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "name",              True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   "(",                 "parenthesis"),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("list",       "inherit_from",      True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
        ("constant",   ")",                 "parenthesis"),  # noqa
        ("formatting", "fifth_formatting",  True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "sixth_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
        ("bool",       "parenthesis",       False),  # noqa
    ],

    "repr": [
        ("constant",   "`",                 True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   "`",                 True),  # noqa
    ],

    "list": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "[",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("list",       "value",             True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   "]",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "associative_parenthesis": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "(",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "value",             True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   ")",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "tuple": [
        ("formatting", "first_formatting",  "with_parenthesis"),  # noqa
        ("constant",   "(",                 "with_parenthesis"),  # noqa
        ("formatting", "second_formatting", "with_parenthesis"),  # noqa
        ("list",       "value",             True),  # noqa
        ("formatting", "third_formatting",  "with_parenthesis"),  # noqa
        ("constant",   ")",                 "with_parenthesis"),  # noqa
        ("formatting", "fourth_formatting", "with_parenthesis"),  # noqa
        ("bool",       "with_parenthesis",  False),  # noqa
    ],

    "await": [
        ("constant",    "await",        True),  # noqa
        ("formatting",  "formatting",   True),  # noqa
        ("key",         "value",        True),  # noqa
    ],

    "def": [
        ("list",       "decorators",        True),  # noqa
        ("bool",       "async",             False),  # noqa
        ("constant",   "async",             "async"),  # noqa
        ("formatting", "async_formatting",  "async"),  # noqa
        ("constant",   "def",               True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "name",              True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   "(",                 True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("list",       "arguments",         True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
        ("constant",   ")",                 True),  # noqa
        ("formatting", "return_annotation_first_formatting", "return_annotation"),  # noqa
        ("constant",   "->", "return_annotation"),  # noqa
        ("formatting", "return_annotation_second_formatting", "return_annotation"),  # noqa
        ("key",        "return_annotation", "return_annotation"),  # noqa
        ("formatting", "fifth_formatting",  True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "sixth_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
    ],

    "call_argument": [
        ("key",        "target",            "target"),  # noqa
        ("formatting", "first_formatting",  "target"),  # noqa
        ("constant",   "=",                 "target"),  # noqa
        ("formatting", "second_formatting", "target"),  # noqa
        ("key",        "value",             True),  # noqa
    ],

    "def_argument": [
        ("key",        "target",            True),  # noqa
        ("formatting", "annotation_first_formatting", "annotation"),  # noqa
        ("constant",   ":",                 "annotation"),  # noqa
        ("formatting", "annotation_second_formatting", "annotation"),  # noqa
        ("key",        "annotation",        "annotation"),  # noqa
        ("formatting", "first_formatting",  "value"),  # noqa
        ("constant",   "=",                 "value"),  # noqa
        ("formatting", "second_formatting", "value"),  # noqa
        ("key",        "value",             "value"),  # noqa
    ],

    "list_argument": [
        ("constant",   "*",                 True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("key",        "value",             True),  # noqa
        ("formatting", "annotation_first_formatting", "annotation"),  # noqa
        ("constant",   ":",                 "annotation"),  # noqa
        ("formatting", "annotation_second_formatting", "annotation"),  # noqa
        ("key",        "annotation",        "annotation"),  # noqa
    ],

    "kwargs_only_marker": [
        ("constant",   "*",                 True),  # noqa
        ("formatting", "formatting",        True),  # noqa
    ],

    "dict_argument": [
        ("constant",   "**",                True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("key",        "value",             True),  # noqa
        ("formatting", "annotation_first_formatting", "annotation"),  # noqa
        ("constant",   ":",                 "annotation"),  # noqa
        ("formatting", "annotation_second_formatting", "annotation"),  # noqa
        ("key",        "annotation",        "annotation"),  # noqa
    ],

    "return": [
        ("constant",   "return",            True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("key",        "value",             "value"),  # noqa
    ],

    "raise": [
        ("constant",   "raise",             True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "value",             "value"),  # noqa
        ("formatting", "second_formatting", "instance"),  # noqa
        ("string",     "comma_or_from",     "instance"),  # noqa
        ("formatting", "third_formatting",  "instance"),  # noqa
        ("key",        "instance",          "instance"),  # noqa
        ("formatting", "fourth_formatting", "traceback"),  # noqa
        ("constant",   ",",                 "traceback"),  # noqa
        ("formatting", "fifth_formatting",  "traceback"),  # noqa
        ("key",        "traceback",         "traceback"),  # noqa
    ],

    "assert": [
        ("constant",   "assert",            True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "value",             True),  # noqa
        ("formatting", "second_formatting", "message"),  # noqa
        ("constant",   ",",                 "message"),  # noqa
        ("formatting", "third_formatting",  "message"),  # noqa
        ("key",        "message",           "message"),  # noqa
    ],

    "set_comprehension": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "{",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "result",            True),  # noqa
        ("list",       "generators",        True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   "}",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "dict_comprehension": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "{",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "result",            True),  # noqa
        ("list",       "generators",        True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   "}",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "argument_generator_comprehension": [
        ("key",  "result",                  True),  # noqa
        ("list", "generators",              True),  # noqa
    ],

    "generator_comprehension": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "(",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "result",            True),  # noqa
        ("list",       "generators",        True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   ")",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "list_comprehension": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "[",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "result",            True),  # noqa
        ("list",       "generators",        True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   "]",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "comprehension_loop": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "for",               True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "iterator",          True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   "in",                True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
        ("key",        "target",            True),  # noqa
        ("list",       "ifs",               True),  # noqa
    ],

    "comprehension_if": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "if",                True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "value",             True),  # noqa
    ],

    "getitem": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "[",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "value",             True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   "]",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "slice": [
        ("key",        "lower",             "lower"),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "upper",             "upper"),  # noqa
        ("formatting", "third_formatting",  "has_two_colons"),  # noqa
        ("constant",   ":",                 "has_two_colons"),  # noqa
        ("formatting", "fourth_formatting", "has_two_colons"),  # noqa
        ("key",        "step",              ["has_two_colons", "step"]),  # noqa
        ("bool",       "has_two_colons",    False),  # noqa
    ],

    "assignment": [
        ("key",        "target",            True),  # noqa
        ("formatting", "annotation_first_formatting",  "annotation"),  # noqa
        ("constant",   ":",                 "annotation"),  # noqa
        ("formatting", "annotation_second_formatting",  "annotation"),  # noqa
        ("key",        "annotation",        "annotation"),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        # FIXME should probably be a different node type  # noqa
        ("string",     "operator",          "operator"),  # noqa
        ("constant",   "=",                 "target"),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "value",             True),  # noqa
    ],

    "standalone_annotation": [
        ("key",        "target",            True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "annotation",        True),  # noqa
    ],

    "unitary_operator": [
        ("string",     "value",             True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("key",        "target",            True),  # noqa
    ],

    "binary_operator": [
        ("key",        "first",             True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "second",            True),  # noqa
    ],

    "boolean_operator": [
        ("key",        "first",             True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("string",     "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "second",            True),  # noqa
    ],

    "comparison_operator": [
        ("string",     "first",             True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("string",     "second",            "second"),  # noqa
    ],

    "comparison": [
        ("key",        "first",             True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "second",            True),  # noqa
    ],

    "with": [
        ("bool",       "async",             False),  # noqa
        ("constant",   "async",             "async"),  # noqa
        ("formatting", "async_formatting",  "async"),  # noqa
        ("constant",   "with",              True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("list",       "contexts",          True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
    ],

    "with_context_item": [
        ("key",        "value",             True),  # noqa
        ("formatting", "first_formatting",  "as"),  # noqa
        ("constant",   "as",                "as"),  # noqa
        ("formatting", "second_formatting", "as"),  # noqa
        ("key",        "as",                "as"),  # noqa
    ],

    "nonlocal": [
        ("constant",    "nonlocal",         True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("list",        "value",            True),  # noqa
    ],

    "del": [
        ("constant",   "del",               True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("key",        "value",             True),  # noqa
    ],

    "yield": [
        ("constant",   "yield",             True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("key",        "value",             "value"),  # noqa
    ],

    "yield_from": [
        ("constant",   "yield",             True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("constant",   "from",              True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "value",             "value"),  # noqa
    ],

    "yield_atom": [
        ("constant",   "(",                 True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "yield",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "value",             "value"),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   ")",                 True),  # noqa
    ],

    "exec": [
        ("constant",   "exec",              True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "value",             True),  # noqa
        ("formatting", "second_formatting", "globals"),  # noqa
        ("constant",   "in",                "globals"),  # noqa
        ("formatting", "third_formatting",  "globals"),  # noqa
        ("key",        "globals",           "globals"),  # noqa
        ("formatting", "fourth_formatting", "locals"),  # noqa
        ("constant",   ",",                 "locals"),  # noqa
        ("formatting", "fifth_formatting",  "locals"),  # noqa
        ("key",        "locals",            "locals"),  # noqa
    ],

    "global": [
        ("constant",   "global",            True),  # noqa
        ("formatting", "formatting",        True),  # noqa
        ("list",       "value",             True),  # noqa
    ],

    "while": [
        ("constant",   "while",             True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "test",              True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
        ("key",        "else",              "else"),  # noqa
    ],

    "for": [
        ("bool",       "async",             False),  # noqa
        ("constant",   "async",             "async"),  # noqa
        ("formatting", "async_formatting",  "async"),  # noqa
        ("constant",   "for",               True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "iterator",          True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   "in",                True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("key",        "target",            True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "fifth_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
        ("key",        "else",              "else"),  # noqa
    ],

    "if": [
        ("constant",   "if",                True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "test",              True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
    ],

    "elif": [
        ("constant",   "elif",              True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "test",              True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
    ],

    "else": [
        ("constant",   "else",              True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("list",       "value",             True),  # noqa
    ],

    "lambda": [
        ("constant",   "lambda",            True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("list",       "arguments",         True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("key",        "value",             True),  # noqa
    ],

    "try": [
        ("constant",   "try",               True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("list",       "value",             True),  # noqa
        ("list",       "excepts",           True),  # noqa
        ("key",        "else",              "else"),  # noqa
        ("key",        "finally",           "finally"),  # noqa
    ],

    "except": [
        ("constant",   "except",            True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("key",        "exception",         "exception"),  # noqa
        ("formatting", "second_formatting", "delimiter"),  # noqa
        ("string",     "delimiter",         "delimiter"),  # noqa
        ("formatting", "third_formatting",  "delimiter"),  # noqa
        ("key",        "target",            "delimiter"),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "fifth_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
    ],

    "finally": [
        ("constant",   "finally",           True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("list",       "value",             True),  # noqa
    ],

    "dict": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "{",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("list",       "value",             True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   "}",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "set": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "{",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("list",       "value",             True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("constant",   "}",                 True),  # noqa
        ("formatting", "fourth_formatting", True),  # noqa
    ],

    "dictitem": [
        ("key",        "key",               True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   ":",                 True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("key",        "value",             True),  # noqa
    ],

    "import": [
        ("formatting", "first_formatting",  True),  # noqa
        ("constant",   "import",            True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("list",       "value",             True),  # noqa
    ],

    "from_import": [
        ("constant",   "from",              True),  # noqa
        ("formatting", "first_formatting",  True),  # noqa
        ("list",       "value",             True),  # noqa
        ("formatting", "second_formatting", True),  # noqa
        ("constant",   "import",            True),  # noqa
        ("formatting", "third_formatting",  True),  # noqa
        ("list",       "targets",           True),  # noqa
    ],

    "dotted_as_name": [
        ("list",       "value",             True),  # noqa
        ("formatting", "first_formatting",  "target"),  # noqa
        ("constant",   "as",                "target"),  # noqa
        ("formatting", "second_formatting", "target"),  # noqa
        ("string",     "target",            "target"),  # noqa
    ],

    "name_as_name": [
        ("string",     "value",             True),  # noqa
        ("formatting", "first_formatting",  "target"),  # noqa
        ("constant",   "as",                "target"),  # noqa
        ("formatting", "second_formatting", "target"),  # noqa
        ("string",     "target",            "target"),  # noqa
    ],

    "print": [
        ("constant",   "print",                  True),  # noqa
        ("formatting", "formatting",             True),  # noqa
        ("constant",   ">>",                     "destination"),  # noqa
        ("formatting", "destination_formatting", "destination"),  # noqa
        ("key",        "destination",            "destination"),  # noqa
        ("list",       "value",                  "value"),  # noqa
    ],
}


class RenderWalker(object):
    """Inherit me and overload the methods you want.

    When calling walk() on a FST node, this class will traverse all the
    node's subtree by following the recipe given by the `render`
    function for the node and recursively for all its childs. At each
    recipe step, it will call methods that you can override to make a
    specific process.

    For every "node", "key", "list", "formatting" and "constant" childs,
    it will call the `before` method when going down the tree and the
    `after` method when going up. There are also specific
    `before_[node,key,list,formatting,constant]` and
    `after_[node,key,list,formatting,constant]` methods provided for
    convenience.

    The latter are called on specific steps:

    * before_list: called before encountering a list of nodes
    * after_list: called after encountering a list of nodes
    * before_formatting: called before encountering a formatting list
    * after_formatting: called after encountering a formatting list
    * before_node: called before encountering a node
    * after_node: called after encountering a node
    * before_key: called before encountering a key type entry
    * after_key: called after encountering a key type entry
    * before_leaf: called before encountering a leaf of the FST (can be a constant (like "def" in a function definition) or an actual value like the value a name node)
    * after_leaf: called after encountering a leaf of the FST (can be a constant (like "def" in a function definition) or an actual value like the value a name node)

    Every method has the same signature: (self, node, render_pos, render_key).
    """
    STOP = True

    def __init__(self, strict=False):
        self.strict = strict

    def before_list(self, node, render_key):
        pass

    def after_list(self, node, render_key):
        pass

    def before_formatting(self, node, render_key):
        pass

    def after_formatting(self, node, render_key):
        pass

    def before_node(self, node, render_key):
        pass

    def after_node(self, node, render_key):
        pass

    def before_key(self, node, render_key):
        pass

    def after_key(self, node, render_key):
        pass

    def before_constant(self, node, render_key):
        pass

    def after_constant(self, node, render_key):
        pass

    def before_string(self, node, render_key):
        pass

    def after_string(self, node, render_key):
        pass

    def before(self, key_type, item, render_key):
        if key_type not in node_types:
            raise NotImplementedError("Unknown key type: %s" % key_type)

        to_call = getattr(self, 'before_' + key_type)

        return to_call(item, render_key)

    def after(self, key_type, item, render_key):
        if key_type not in node_types:
            raise NotImplementedError("Unknown key type: %s" % key_type)

        to_call = getattr(self, 'after_' + key_type)

        return to_call(item, render_key)

    def walk(self, node):
        return self._walk(node)

    def _walk(self, node):
        for key_type, item, render_key in render(node, strict=getattr(self, "strict", False)):
            stop = self._walk_on_item(key_type, item, render_key)
            if stop == self.STOP:
                return self.STOP

    def _walk_on_item(self, key_type, item, render_key):
        stop_before = self.before(key_type, item, render_key)
        if stop_before:
            return self.STOP

        stop = self._walk(item) if key_type not in ['constant', 'string'] else False

        stop_after = self.after(key_type, item, render_key)

        if stop or stop_after:
            return self.STOP



# encoding: utf-8

import re
from .utils import FlexibleIterator

to_group = (
    ("+", "="),
    ("-", "="),
    ("*", "="),
    ("/", "="),
    ("%", "="),
    ("&", "="),
    ("|", "="),
    ("^", "="),
    ("@", "="),
    ("/", "/"),
    ("*", "*"),
    ("<", "<"),
    (">", ">"),
    ("=", "="),
    ("!", "="),
    ("<", ">"),
    ("<", "="),
    (">", "="),
    ("**", "="),
    ("//", "="),
    ("<<", "="),
    (">>", "="),
    ("\r", "\n"),
    (".", "."),
    ("..", "."),
    ("-", ">"),
)

to_group_keys, _ = list(zip(*to_group))


def group(sequence):
    return list(group_generator(sequence))


def match_on_next(regex, iterator):
    return iterator.show_next() and re.match(regex, iterator.show_next())


def group_generator(sequence):
    iterator = FlexibleIterator(sequence)
    current = None
    while True:
        if iterator.end():
            return

        current = next(iterator)
        if current in to_group_keys and matching_found(to_group, current, iterator.show_next()):
            current += next(iterator)
        if current in to_group_keys and matching_found(to_group, current, iterator.show_next()):
            current += next(iterator)
        if current in list('uUfFrRbB') and str(iterator.show_next()).startswith(('"', "'")):
            current += next(iterator)
        if str(current).lower() in ["ur", "br", "fr", "rf"] and str(iterator.show_next()).startswith(('"', "'")):
            current += next(iterator)
        if any([re.match(x, current) for x in (r'^\d+[eE]$', r'^\d+\.\d*[eE]$', r'^\.\d+[eE]$')]):
            current += next(iterator)
            current += next(iterator)

            # It's required in a case where I have something like that:
            # ['123.123e', '[+-]', '123']
            assert re.match(r'^\d+[eE][-+]?\d+[jJ]?$', current) or re.match(r'^\d*.\d*[eE][-+]?\d+[jJ]?$', current)

        if current == "\\" and iterator.show_next() in ('\n', '\r\n'):
            current += next(iterator)
            if re.match(r'^\s+$', str(iterator.show_next())):
                current += next(iterator)

        if current == "\\" and iterator.show_next() == "\r" and iterator.show_next(2) == "\n":
            current += next(iterator)
            current += next(iterator)
            if re.match(r'^\s+$', str(iterator.show_next())):
                current += next(iterator)

        if re.match(r'^\s+$', current) and iterator.show_next() == "\\":
            current += next(iterator)
            current += next(iterator)
            if iterator.show_next() == "\n":
                current += next(iterator)
            if re.match(r'^\s+$', str(iterator.show_next())):
                current += next(iterator)

        if (re.match(r'^[_\d]+$', current) and match_on_next(r'^\.$', iterator)) or\
           (current == "." and match_on_next(r'^\d+[_\d]*([jJ]|[eE]\d*)?$', iterator)):
            current += next(iterator)

            if match_on_next(r'^[_\d]*[jJ]?$', iterator) and match_on_next(r'^[_\d]*[jJ]?$', iterator).group():
                current += next(iterator)

        if re.match(r'^\d+\.$', current) and match_on_next(r'^\d*[eE]\d*$', iterator):
            current += next(iterator)

        if re.match(r'^\d+\.?[eE]$', current) and match_on_next(r'^\d+$', iterator):
            current += next(iterator)

        if re.match(r'^\d*\.?\d*[eE]$', current) and not re.match('[eE]', current) and match_on_next(r'^[-+]$', iterator) and iterator.show_next(2) and re.match(r'^\d+$', iterator.show_next(2)):
            current += next(iterator)
            current += next(iterator)

        # edge case where 2 dots follow themselves but not 3 (an ellipsis)
        if current == "..":
            yield "."
            yield "."
            continue

        yield current


def matching_found(to_group, current, target):
    return target in [x[1] for x in to_group if x[0] == current]



# -*- coding: utf-8 -*-

from baron import (parse, BaronError, ParsingError,
                   GroupingError, UntreatedError)
import pytest


def test_dummy_parse():
    parse("pouet")


def test_error_parsing_error():
    with pytest.raises(ParsingError):
        parse("(")
    with pytest.raises(BaronError):
        parse("(")


def test_error_unexpected_formatting():
    with pytest.raises(ParsingError):
        parse("   a\nb")
    with pytest.raises(BaronError):
        parse("   a\nb")


def test_error_grouping():
    with pytest.raises(GroupingError):
        parse("   (a\n b")
    with pytest.raises(BaronError):
        parse("   (a\n b")


def test_error_untreated_error():
    with pytest.raises(UntreatedError):
        parse("?")
    with pytest.raises(BaronError):
        parse("?")


def test_missing_quote_yields_error():
    with pytest.raises(UntreatedError):
        parse("'")
    with pytest.raises(UntreatedError):
        parse("'\n")
    with pytest.raises(BaronError):
        parse("'\n")


def test_error_bad_indentation():
    """ Regression test case

    It shows a discrepency between python2 and python3 in string
    comparisons with None.
    """
    with pytest.raises(ParsingError):
        parse("def fun():\nass")
    with pytest.raises(BaronError):
        parse("def fun():\nass")



#!/usr/bin/python
# -*- coding:Utf-8 -*-
import pytest
from baron.parser import ParsingError
from .test_utils import parse_simple, parse_multi


def test_empty():
    ""
    parse_simple([], [])


def test_int():
    "1"
    parse_simple([('INT', '1')], [
        {
            "type": "int",
            "section": "number",
            "value": "1",
        }
    ])


def test_long():
    "123234L"
    parse_simple([('LONG', '123234L')], [
        {
            "type": "long",
            "section": "number",
            "value": "123234L",
        }
    ])


def test_name():
    "a"
    parse_simple([('NAME', 'a')], [{"type": "name", "value": "a", }])


def test_string():
    '''
    "pouet pouet"
    """pouet pouet"""
    '''
    parse_simple([('STRING', '"pouet pouet"')], [
        {
            "type": "string",
            "value": '"pouet pouet"',
            "first_formatting": [],
            "second_formatting": [],
        }
    ])
    parse_simple([('STRING', '"""pouet pouet"""')], [
        {
            "type": "string",
            "value": '"""pouet pouet"""',
            "first_formatting": [],
            "second_formatting": [],
        }
    ])


def test_file_input_empty():
    ""
    parse_multi([], [])


def test_file_input_one_item():
    "a"
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])


def test_file_input_two_items():
    """
    a
    a
    """
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
    ])
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
    ])
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])


def test_file_input_two_items_endl():
    """
    a

    a
    """
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
    ])
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
    ])
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
    ])
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
    ])
    parse_multi([
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('ENDL', '\n'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])


def test_file_input_simple_stmt_one_item_semicolon():
    """
    a;
    """
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {
            "type": "semicolon",
            "value": ";",
            "first_formatting": [],
            "second_formatting": [],
        },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])


def test_file_input_simple_stmt_two_items_semicolon():
    """
    a;a
    """
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {
            "type": "semicolon",
            "value": ";",
            "first_formatting": [],
            "second_formatting": [],
        },
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {
            "type": "semicolon",
            "value": ";",
            "first_formatting": [],
            "second_formatting": [],
        },
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])


def test_file_input_simple_stmt_three_items_semicolon():
    """
    a;b;a
    """
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';'),
        ('NAME', 'b'),
        ('SEMICOLON', ';'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": []},
        {"type": "name", "value": 'b'},
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": [], },
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';'),
        ('NAME', 'b'),
        ('SEMICOLON', ';'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": []},
        {"type": "name", "value": 'b', },
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": [], },
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';'),
        ('NAME', 'b'),
        ('SEMICOLON', ';'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": []},
        {"type": "name", "value": 'b'},
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": [], },
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';'),
        ('NAME', 'b'),
        ('SEMICOLON', ';'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": []},
        {"type": "name", "value": 'b'},
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": [], },
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';'),
        ('NAME', 'b'),
        ('SEMICOLON', ';'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": []},
        {"type": "name", "value": 'b', },
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": [], },
        {"type": "name", "value": 'a'},
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';'),
        ('NAME', 'b'),
        ('SEMICOLON', ';'),
        ('NAME', 'a'),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a'},
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": []},
        {"type": "name", "value": 'b'},
        {"type": "semicolon", "value": ";", "first_formatting": [], "second_formatting": [], },
        {"type": "name", "value": 'a', },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])


def test_file_input_simple_stmt_one_item_semicolon_space():
    """
    a ;
    """
    parse_multi([
        ('NAME', 'a'),
        ('SEMICOLON', ';', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('ENDL', '\n'),
    ], [
        {"type": "name", "value": 'a', },
        {"type": "semicolon", "value": ";", "first_formatting": [{"type": "space", "value": ' '}], "second_formatting": [{"type": "space", "value": ' '}], },
        {"type": "endl", "value": "\n", "formatting": [], "indent": "", },
    ])


def test_ellipsis():
    "..."
    parse_simple([('ELLIPSIS', '...')], [
        {
            "type": "ellipsis",
            "first_formatting": [],
            "second_formatting": [],
        }
    ])


def test_funcdef_stmt_indent():
    """
    def a () :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')]),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "arguments": [],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_indent_return_annotation():
    """
    def a () -> b :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')]),
        ('RIGHT_PARENTHESIS', ')'),
        ('RIGHT_ARROW', '->', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COLON', ':', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "return_annotation": {"type": "name", "value": "b"},
            "return_annotation_first_formatting": [{"type": "space", "value": " "}],
            "return_annotation_second_formatting": [{"type": "space", "value": " "}],
            "async_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "arguments": [],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_async():
    """
    async def a():
        pass
    """
    parse_multi([
        ('NAME', 'async', [], []),
        ('SPACE', ' '),
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')]),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": True,
            "async_formatting": [{"type": "space", "value": " "}],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "arguments": [],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_async_bad_keyword():
    """
    async def a():
        pass
    """
    with pytest.raises(ParsingError):
        parse_multi([
            ('NAME', 'not_async', [], []),
            ('SPACE', ' '),
            ('DEF', 'def', [], [('SPACE', ' ')]),
            ('NAME', 'a'),
            ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')]),
            ('RIGHT_PARENTHESIS', ')'),
            ('COLON', ':', [], [('SPACE', ' ')]),
            ('ENDL', '\n', [], [('SPACE', '    ')]),
            ('INDENT', ''),
            ('PASS', 'pass'),
            ('ENDL', '\n'),
            ('DEDENT', ''),
        ], [])


def test_funcdef_stmt_one_parameter_indent():
    """
    def a ( x ) :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'x'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "arguments": [
                {
                    "type": "def_argument",
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "first_formatting": [],
                    "second_formatting": [],
                    "target": {
                        "type": "name",
                        "value": "x",
                    },
                    "value": {},
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_one_parameter_typed_indent():
    """
    def a ( x : int ) :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'x'),
        ('COLON', ':', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'int'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "arguments": [
                {
                    "type": "def_argument",
                    "first_formatting": [],
                    "second_formatting": [],
                    "annotation": {"type": "name", "value": "int"},
                    "annotation_first_formatting": [{"type": "space", "value": " "}],
                    "annotation_second_formatting": [{"type": "space", "value": " "}],
                    "target": {
                        "type": "name",
                        "value": "x",
                    },
                    "value": {},
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_one_parameter_typed_indent_no_format():
    """
    def a ( x:int ) :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'x'),
        ('COLON', ':', [], []),
        ('NAME', 'int'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "arguments": [
                {
                    "type": "def_argument",
                    "annotation": {"type": "name", "value": "int"},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "first_formatting": [],
                    "second_formatting": [],
                    "target": {
                        "type": "name",
                        "value": "x",
                    },
                    "value": {},
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_star_parameter_typed_indent_no_format():
    """
    def a ( *x:int ) :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('STAR', 'x'),
        ('NAME', 'x'),
        ('COLON', ':', [], []),
        ('NAME', 'int'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "arguments": [
                {
                    "type": "list_argument",
                    "annotation": {"type": "name", "value": "int"},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "formatting": [],
                    "value": {
                        "type": "name",
                        "value": "x",
                    },
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_double_star_parameter_typed_indent_no_format():
    """
    def a ( **x:int ) :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('DOUBLE_STAR', '**'),
        ('NAME', 'x'),
        ('COLON', ':', [], []),
        ('NAME', 'int'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "arguments": [
                {
                    "type": "dict_argument",
                    "annotation": {"type": "name", "value": "int"},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "formatting": [],
                    "value": {
                        "type": "name",
                        "value": "x",
                    },
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_one_parameter_comma_indent():
    """
    def a ( x , ) :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'x'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "name": "a",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "arguments": [
                {
                    "type": "def_argument",
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "first_formatting": [],
                    "second_formatting": [],
                    "target": {
                        "type": "name",
                        "value": "x",
                    },
                    "value": {},
                },
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_one_parameter_comma_default_indent():
    """
    def a ( x=1 , ) :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'x'),
        ('EQUAL', '='),
        ('INT', '1'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "name": "a",
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "arguments": [
                {
                    "type": "def_argument",
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "first_formatting": [],
                    "second_formatting": [],
                    "target": {
                        "type": "name",
                        "value": "x",
                    },
                    "value": {
                        "type": "int",
                        "value": "1",
                        "section": "number",
                    },
                },
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_two_parameters_typed_with_default_indent():
    """
    def a ( x : int = 1 , y : List[str] ) :
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'x'),
        ('COLON', ':', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'int'),
        ('EQUAL', '=', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('INT', '1'),
        ('COMMA', ',', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'y'),
        ('COLON', ':', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'List'),
        ('LEFT_SQUARE_BRACKET', '[', [], []),
        ('NAME', 'str'),
        ('RIGHT_SQUARE_BRACKET', '[', [], []),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')]),
        ('COLON', ':', [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "arguments": [
                {
                    "type": "def_argument",
                    "annotation": {"type": "name", "value": "int"},
                    "annotation_first_formatting": [{"type": "space", "value": " "}],
                    "annotation_second_formatting": [{"type": "space", "value": " "}],
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "target": {
                        "type": "name",
                        "value": "x",
                    },
                    "value": {"section": "number", "type": "int", "value": "1"},
                },
                {
                    "type": "comma",
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "def_argument",
                    "annotation": {
                        "type": "atomtrailers",
                        "value": [
                            {
                                "type": "name",
                                "value": "List",
                            },
                            {
                                "type": "getitem",
                                "first_formatting": [],
                                "second_formatting": [],
                                "third_formatting": [],
                                "fourth_formatting": [],
                                "value": {"type": "name", "value": "str"},
                            }
                        ]
                    },
                    "annotation_first_formatting": [{"type": "space", "value": " "}],
                    "annotation_second_formatting": [{"type": "space", "value": " "}],
                    "first_formatting": [],
                    "second_formatting": [],
                    "target": {
                        "type": "name",
                        "value": "y",
                    },
                    "value": {},
                },
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "type": "endl",
                    "formatting": [],
                    "indent": "",
                    "value": "\n"
                }
            ],
        }
    ])


def test_class_empty():
    """
    class A:
        pass
    """
    parse_multi([
        ('CLASS', 'class', [], [('SPACE', ' ')]),
        ('NAME', 'A'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n', [], []),
        ('DEDENT', ''),
    ], [
        {
            "type": "class",
            "name": "A",
            "decorators": [],
            "parenthesis": False,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "sixth_formatting": [],
            "inherit_from": [],
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_class_empty_parenthesis():
    """
    class A ( ) :
        pass
    """
    parse_multi([
        ('CLASS', 'class', [], [('SPACE', ' ')]),
        ('NAME', 'A'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('RIGHT_PARENTHESIS', ')', [], [('SPACE', ' ')]),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n', [], []),
        ('DEDENT', ''),
    ], [
        {
            "type": "class",
            "name": "A",
            "decorators": [],
            "parenthesis": True,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "inherit_from": [],
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_class_inherit():
    """
    class A ( B ) :
        pass
    """
    parse_multi([
        ('CLASS', 'class', [], [('SPACE', ' ')]),
        ('NAME', 'A'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'B'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n', [], []),
        ('DEDENT', ''),
    ], [
        {
            "type": "class",
            "name": "A",
            "decorators": [],
            "parenthesis": True,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "inherit_from": [
                {
                    "type": "name",
                    "value": "B"
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_class_inherit_metaclass():
    """
    class A ( B = C ) :
        pass
    """
    parse_multi([
        ('CLASS', 'class', [], [('SPACE', ' ')]),
        ('NAME', 'A'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'B'),
        ('EQUAL', '=', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'C'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n', [], []),
        ('DEDENT', ''),
    ], [
        {
            "type": "class",
            "name": "A",
            "decorators": [],
            "parenthesis": True,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "inherit_from": [
                {
                    "first_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "second_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "target": {
                        "type": "name",
                        "value": "B"
                    },
                    "type": "call_argument",
                    "value": {
                        "type": "name",
                        "value": "C"
                    }
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_class_inherit_metaclass_arglist():
    """
    class A ( B, C = D ) :
        pass
    """
    parse_multi([
        ('CLASS', 'class', [], [('SPACE', ' ')]),
        ('NAME', 'A'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'B'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'C'),
        ('EQUAL', '=', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'D'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n', [], []),
        ('DEDENT', ''),
    ], [
        {
            "type": "class",
            "name": "A",
            "decorators": [],
            "parenthesis": True,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "inherit_from": [
                {
                    "type": "name",
                    "value": "B"
                },
                {
                    "first_formatting": [],
                    "second_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "type": "comma"
                },
                {
                    "first_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "second_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "target": {
                        "type": "name",
                        "value": "C"
                    },
                    "type": "call_argument",
                    "value": {
                        "type": "name",
                        "value": "D"
                    }
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_class_inherit_metaclass_arglist_more():
    """
    class A ( B, E, F, C = D ) :
        pass
    """
    parse_multi([
        ('CLASS', 'class', [], [('SPACE', ' ')]),
        ('NAME', 'A'),
        ('LEFT_PARENTHESIS', '(', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'B'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'E'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'F'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'C'),
        ('EQUAL', '=', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'D'),
        ('RIGHT_PARENTHESIS', ')', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n', [], []),
        ('DEDENT', ''),
    ], [
        {
            "type": "class",
            "name": "A",
            "decorators": [],
            "parenthesis": True,
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "third_formatting": [{"type": "space", "value": " "}],
            "fourth_formatting": [{"type": "space", "value": " "}],
            "fifth_formatting": [{"type": "space", "value": " "}],
            "sixth_formatting": [],
            "inherit_from": [
                {
                    "type": "name",
                    "value": "B"
                },
                {
                    "first_formatting": [],
                    "second_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "type": "comma"
                },
                {
                    "type": "name",
                    "value": "E"
                },
                {
                    "first_formatting": [],
                    "second_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "type": "comma"
                },
                {
                    "type": "name",
                    "value": "F"
                },
                {
                    "first_formatting": [],
                    "second_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "type": "comma"
                },
                {
                    "first_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "second_formatting": [
                        {
                            "type": "space",
                            "value": " "
                        }
                    ],
                    "target": {
                        "type": "name",
                        "value": "C"
                    },
                    "type": "call_argument",
                    "value": {
                        "type": "name",
                        "value": "D"
                    }
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_one_start_parameter_indent():
    """
    def a (*b):
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '('),
        ('STAR', '*'),
        ('NAME', 'b'),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "arguments": [
                {
                    "type": "list_argument",
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "formatting": [],
                    "value": {
                        "value": "b",
                        "type": "name",
                    }
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "value": "\n",
                    "formatting": [],
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_funcdef_stmt_one_star_star_parameter_indent():
    """
    def a (**b):
        pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '('),
        ('DOUBLE_STAR', '**'),
        ('NAME', 'b'),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
    ], [
        {
            "type": "def",
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "arguments": [
                {
                    "type": "dict_argument",
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "formatting": [],
                    "value": {
                        "value": "b",
                        "type": "name",
                    }
                }
            ],
            "value": [
                {
                    "type": "endl",
                    "formatting": [],
                    "value": "\n",
                    "indent": "    "
                },
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n"
                }
            ],
        }
    ])


def test_comment():
    """
      # comment
    """
    parse_multi([
        ('COMMENT', '# comment', [('SPACE', '  ')]),
        ('ENDL', '\n'),
    ], [
        {
            "formatting": [{"type": "space", "value": "  "}],
            "type": "comment",
            "value": "# comment",
        },
        {
            "formatting": [],
            "indent": "",
            "type": "endl",
            "value": "\n",
        }
    ])


def test_comment_consistant():
    """
      # comment
    """
    parse_multi([
        ('COMMENT', '# comment'),
        ('ENDL', '\n'),
    ], [
        {
            "formatting": [],
            "type": "comment",
            "value": "# comment",
        },
        {
            "formatting": [],
            "indent": "",
            "type": "endl",
            "value": "\n",
        }
    ])


def test_with_a():
    """
    with a: pass
    """
    parse_multi([
        ('WITH', 'with', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "type": "with",
            "async": False,
            "async_formatting": [],
            "contexts": [
                {
                    "type": "with_context_item",
                    "value": {
                        "type": "name",
                        "value": "a",
                    },
                    "first_formatting": [],
                    "second_formatting": [],
                    "as": {},
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_with_a_as_b():
    """
    with a as b: pass
    """
    parse_multi([
        ('WITH', 'with', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('AS', 'as', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "type": "with",
            "async": False,
            "async_formatting": [],
            "contexts": [
                {
                    "type": "with_context_item",
                    "value": {
                        "type": "name",
                        "value": "a",
                    },
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "as": {
                        "type": "name",
                        "value": "b",
                    },
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_with_a_as_b_c():
    """
    with a as b, c: pass
    """
    parse_multi([
        ('WITH', 'with', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('AS', 'as', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "type": "with",
            "async": False,
            "async_formatting": [],
            "contexts": [
                {
                    "type": "with_context_item",
                    "value": {
                        "type": "name",
                        "value": "a",
                    },
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "as": {
                        "type": "name",
                        "value": "b",
                    },
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "with_context_item",
                    "value": {
                        "type": "name",
                        "value": "c",
                    },
                    "first_formatting": [],
                    "second_formatting": [],
                    "as": {},
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_async_with_a():
    """
    async with a: pass
    """
    parse_multi([
        ('NAME', 'async', [], []),
        ('SPACE', ' '),
        ('WITH', 'with', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "type": "with",
            "async": True,
            "async_formatting": [{"type": "space", "value": " "}],
            "contexts": [
                {
                    "type": "with_context_item",
                    "value": {
                        "type": "name",
                        "value": "a",
                    },
                    "first_formatting": [],
                    "second_formatting": [],
                    "as": {},
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_async_with_a_as_b():
    """
    async with a as b: pass
    """
    parse_multi([
        ('NAME', 'async', [], []),
        ('SPACE', ' '),
        ('WITH', 'with', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('AS', 'as', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "type": "with",
            "async": True,
            "async_formatting": [{"type": "space", "value": " "}],
            "contexts": [
                {
                    "type": "with_context_item",
                    "value": {
                        "type": "name",
                        "value": "a",
                    },
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "as": {
                        "type": "name",
                        "value": "b",
                    },
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_async_with_a_as_b_c():
    """
    async with a as b, c: pass
    """
    parse_multi([
        ('NAME', 'async', [], []),
        ('SPACE', ' '),
        ('WITH', 'with', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('AS', 'as', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('COMMA', ',', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [{"type": "space", "value": " "}],
            "type": "with",
            "async": True,
            "async_formatting": [{"type": "space", "value": " "}],
            "contexts": [
                {
                    "type": "with_context_item",
                    "value": {
                        "type": "name",
                        "value": "a",
                    },
                    "first_formatting": [{"type": "space", "value": " "}],
                    "second_formatting": [{"type": "space", "value": " "}],
                    "as": {
                        "type": "name",
                        "value": "b",
                    },
                },
                {
                    "type": "comma",
                    "first_formatting": [],
                    "second_formatting": [{"type": "space", "value": " "}],
                },
                {
                    "type": "with_context_item",
                    "value": {
                        "type": "name",
                        "value": "c",
                    },
                    "first_formatting": [],
                    "second_formatting": [],
                    "as": {},
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_async_with_a_bad_keyword():
    """
    async with a: pass
    """
    with pytest.raises(ParsingError):
        parse_multi([
            ('NAME', 'not_async', [], []),
            ('SPACE', ' '),
            ('WITH', 'with', [], [('SPACE', ' ')]),
            ('NAME', 'a'),
            ('COLON', ':', [], [('SPACE', ' ')]),
            ('PASS', 'pass'),
            ('ENDL', '\n'),
        ], [])


def test_async_with_a_as_b_bad_keyword():
    """
    async with a as b: pass
    """
    with pytest.raises(ParsingError):
        parse_multi([
            ('NAME', 'not_async', [], []),
            ('SPACE', ' '),
            ('WITH', 'with', [], [('SPACE', ' ')]),
            ('NAME', 'a'),
            ('AS', 'as', [('SPACE', ' ')], [('SPACE', ' ')]),
            ('NAME', 'b'),
            ('COLON', ':', [], [('SPACE', ' ')]),
            ('PASS', 'pass'),
            ('ENDL', '\n'),
        ], [])


def test_async_with_a_as_b_c_bad_keyword():
    """
    async with a as b, c: pass
    """
    with pytest.raises(ParsingError):
        parse_multi([
            ('NAME', 'not_async', [], []),
            ('SPACE', ' '),
            ('WITH', 'with', [], [('SPACE', ' ')]),
            ('NAME', 'a'),
            ('AS', 'as', [('SPACE', ' ')], [('SPACE', ' ')]),
            ('NAME', 'b'),
            ('COMMA', ',', [], [('SPACE', ' ')]),
            ('NAME', 'c'),
            ('COLON', ':', [], [('SPACE', ' ')]),
            ('PASS', 'pass'),
            ('ENDL', '\n'),
        ], [])


def test_decorator():
    """
    @a
    def b(): pass
    """
    parse_multi([
        ('AT', '@', [], []),
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('LEFT_PARENTHESIS', '('),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fifth_formatting": [],
            "fourth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "type": "def",
            "arguments": [],
            "name": "b",
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "decorators": [
                {
                    "type": "decorator",
                    "formatting": [],
                    "call": {},
                    "value": {
                        "type": "dotted_name",
                        "value": [
                            {
                                "type": "name",
                                "value": "a",
                            }
                        ],
                    }
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_decorator_parenthesis():
    """
    @a()
    def b(): pass
    """
    parse_multi([
        ('AT', '@', [], []),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '('),
        ('RIGHT_PARENTHESIS', ')'),
        ('ENDL', '\n'),
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('LEFT_PARENTHESIS', '('),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fifth_formatting": [],
            "fourth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "type": "def",
            "arguments": [],
            "name": "b",
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "decorators": [
                {
                    "type": "decorator",
                    "formatting": [],
                    "call": {
                        "third_formatting": [],
                        "type": "call",
                        "first_formatting": [],
                        "value": [],
                        "second_formatting": [],
                        "fourth_formatting": [],
                    },
                    "value": {
                        "type": "dotted_name",
                        "value": [
                            {
                                "type": "name",
                                "value": "a",
                            }
                        ],
                    }
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_decorator_parenthesis_arg():
    """
    @a(c)
    def b(): pass
    """
    parse_multi([
        ('AT', '@', [], []),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '('),
        ('NAME', 'c'),
        ('RIGHT_PARENTHESIS', ')'),
        ('ENDL', '\n'),
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('LEFT_PARENTHESIS', '('),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fifth_formatting": [],
            "fourth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "type": "def",
            "arguments": [],
            "name": "b",
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "decorators": [
                {
                    "type": "decorator",
                    "formatting": [],
                    "call": {
                        "third_formatting": [],
                        "type": "call",
                        "fourth_formatting": [],
                        "first_formatting": [],
                        "value": [
                            {
                                "target": {},
                                "first_formatting": [],
                                "second_formatting": [],
                                "type": "call_argument",
                                "value": {
                                    "type": "name",
                                    "value": "c",
                                },
                            }
                        ],
                        "second_formatting": [],
                    },
                    "value": {
                        "type": "dotted_name",
                        "value": [
                            {
                                "type": "name",
                                "value": "a",
                            }
                        ],
                    }
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_decorator_two():
    """
    @a
    @ c
    def b(): pass
    """
    parse_multi([
        ('AT', '@', [], []),
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('AT', '@', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('ENDL', '\n'),
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('LEFT_PARENTHESIS', '('),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fifth_formatting": [],
            "fourth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "type": "def",
            "arguments": [],
            "name": "b",
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "decorators": [
                {
                    "type": "decorator",
                    "formatting": [],
                    "call": {},
                    "value": {
                        "type": "dotted_name",
                        "value": [
                            {
                                "type": "name",
                                "value": "a",
                            }
                        ],
                    }
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                },
                {
                    "type": "decorator",
                    "formatting": [{"type": "space", "value": " "}],
                    "call": {},
                    "value": {
                        "type": "dotted_name",
                        "value": [
                            {
                                "type": "name",
                                "value": "c",
                            }
                        ],
                    }
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_class_decorator():
    """
    @a
    class b(): pass
    """
    parse_multi([
        ('AT', '@', [], []),
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('CLASS', 'class', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('LEFT_PARENTHESIS', '('),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fifth_formatting": [],
            "fourth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "type": "class",
            "inherit_from": [],
            "parenthesis": True,
            "name": "b",
            "decorators": [
                {
                    "type": "decorator",
                    "call": {},
                    "formatting": [],
                    "value": {
                        "type": "dotted_name",
                        "value": [
                            {
                                "type": "name",
                                "value": "a",
                            }
                        ],
                    }
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_class_decorator_two():
    """
    @a
    @ c
    class b(): pass
    """
    parse_multi([
        ('AT', '@', [], []),
        ('NAME', 'a'),
        ('ENDL', '\n'),
        ('AT', '@', [], [('SPACE', ' ')]),
        ('NAME', 'c'),
        ('ENDL', '\n'),
        ('CLASS', 'class', [], [('SPACE', ' ')]),
        ('NAME', 'b'),
        ('LEFT_PARENTHESIS', '('),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fifth_formatting": [],
            "fourth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "type": "class",
            "inherit_from": [],
            "parenthesis": True,
            "name": "b",
            "decorators": [
                {
                    "type": "decorator",
                    "call": {},
                    "formatting": [],
                    "value": {
                        "type": "dotted_name",
                        "value": [
                            {
                                "type": "name",
                                "value": "a",
                            }
                        ],
                    }
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                },
                {
                    "type": "decorator",
                    "call": {},
                    "formatting": [{"type": "space", "value": " "}],
                    "value": {
                        "type": "dotted_name",
                        "value": [
                            {
                                "type": "name",
                                "value": "c",
                            }
                        ],
                    }
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_fplist():
    """
    def a((b,)): pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '('),
        ('LEFT_PARENTHESIS', '('),
        ('NAME', 'b'),
        ('COMMA', ','),
        ('RIGHT_PARENTHESIS', ')'),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fifth_formatting": [],
            "fourth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "type": "def",
            "arguments": [
                {
                    "type": "def_argument",
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "first_formatting": [],
                    "second_formatting": [],
                    "value": {},
                    "target": {
                        "first_formatting": [],
                        "second_formatting": [],
                        "third_formatting": [],
                        "fourth_formatting": [],
                        "type": "tuple",
                        "with_parenthesis": True,
                        "value": [
                            {
                                "value": "b",
                                "type": "name",
                            },
                            {
                                "first_formatting": [],
                                "second_formatting": [],
                                "type": "comma",
                            }
                        ],
                    },
                }
            ],
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_fplist_two():
    """
    def a((b,c)): pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '('),
        ('LEFT_PARENTHESIS', '('),
        ('NAME', 'b'),
        ('COMMA', ','),
        ('NAME', 'c'),
        ('RIGHT_PARENTHESIS', ')'),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fifth_formatting": [],
            "fourth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "type": "def",
            "arguments": [
                {
                    "type": "def_argument",
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "first_formatting": [],
                    "second_formatting": [],
                    "value": {},
                    "target": {
                        "type": "tuple",
                        "with_parenthesis": True,
                        "first_formatting": [],
                        "second_formatting": [],
                        "third_formatting": [],
                        "fourth_formatting": [],
                        "value": [
                            {
                                "type": "name",
                                "value": "b",
                            },
                            {
                                "type": "comma",
                                "first_formatting": [],
                                "second_formatting": [],
                            },
                            {
                                "type": "name",
                                "value": "c",
                            }
                        ],
                    }
                }
            ],
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_fplist_alone():
    """
    def a((b)): pass
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'a'),
        ('LEFT_PARENTHESIS', '('),
        ('LEFT_PARENTHESIS', '('),
        ('NAME', 'b'),
        ('RIGHT_PARENTHESIS', ')'),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':', [], [('SPACE', ' ')]),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
    ], [
        {
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [],
            "third_formatting": [],
            "fifth_formatting": [],
            "fourth_formatting": [],
            "sixth_formatting": [{"type": "space", "value": " "}],
            "type": "def",
            "arguments": [
                {
                    "type": "def_argument",
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    "value": {},
                    "first_formatting": [],
                    "second_formatting": [],
                    "target": {
                        "type": "associative_parenthesis",
                        "first_formatting": [],
                        "second_formatting": [],
                        "third_formatting": [],
                        "fourth_formatting": [],
                        "value": {
                            "type": "name",
                            "value": "b",
                        }
                    }
                }
            ],
            "name": "a",
            "decorators": [],
            "async": False,
            "async_formatting": [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            "value": [
                {
                    "type": "pass",
                },
                {
                    "formatting": [],
                    "indent": "",
                    "type": "endl",
                    "value": "\n",
                }
            ],
        }
    ])


def test_endl_dont_grab_comment_as_indent():
    """
    \n# pouet
    """
    parse_multi([
        ('ENDL', '\n', [], [('COMMENT', '# pouet')])
    ], [
        {"type": "endl", "value": "\n", "indent": "", "formatting": []},
        {"type": "comment", "value": "# pouet", "formatting": []},
    ])


def test_regression_def_argument_tuple():
    """
    def function_name((a,b)=c):\n    pass\n
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'function_name'),
        ('LEFT_PARENTHESIS', '('),
        ('LEFT_PARENTHESIS', '('),
        ('NAME', 'a'),
        ('COMMA', ','),
        ('NAME', 'b'),
        ('RIGHT_PARENTHESIS', ')'),
        ('EQUAL', '='),
        ('NAME', 'c'),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n', [], []),
        ('DEDENT', ''),
        ('ENDMARKER', '')
    ], [
        {
            'arguments': [
                {
                    'first_formatting': [],
                    'second_formatting': [],
                    'type': 'def_argument',
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    'target': {
                        'type': 'tuple',
                        'first_formatting': [],
                        'with_parenthesis': True,
                        'fourth_formatting': [],
                        'second_formatting': [],
                        'third_formatting': [],
                        'value': [
                            {
                                'value': 'a',
                                'type': 'name',
                            },
                            {
                                'first_formatting': [],
                                'second_formatting': [],
                                'type': 'comma'
                            },
                            {
                                'value': 'b',
                                'type': 'name',
                            }
                        ]
                    },
                    'value': {
                        'type': 'name',
                        'value': 'c',
                    }
                }
            ],
            'decorators': [],
            'async': False,
            'async_formatting': [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            'fifth_formatting': [],
            'first_formatting': [
                {
                    'type': 'space',
                    'value': ' '
                }
            ],
            'fourth_formatting': [],
            'name': 'function_name',
            'second_formatting': [],
            'sixth_formatting': [],
            'third_formatting': [],
            'type': 'def',
            'value': [
                {
                    'formatting': [],
                    'indent': '    ',
                    'type': 'endl',
                    'value': '\n'
                },
                {
                    'type': 'pass'
                },
                {
                    'formatting': [],
                    'indent': '',
                    'type': 'endl',
                    'value': '\n'
                }
            ]
        }
    ])


def test_regression_def_argument_tuple_nested():
    """
    def function_name((a,(x,y))=c):\n    pass\n
    """
    parse_multi([
        ('DEF', 'def', [], [('SPACE', ' ')]),
        ('NAME', 'function_name'),
        ('LEFT_PARENTHESIS', '('),
        ('LEFT_PARENTHESIS', '('),
        ('NAME', 'a'),
        ('COMMA', ','),
        ('LEFT_PARENTHESIS', '('),
        ('NAME', 'x'),
        ('COMMA', ','),
        ('NAME', 'y'),
        ('RIGHT_PARENTHESIS', ')'),
        ('RIGHT_PARENTHESIS', ')'),
        ('EQUAL', '='),
        ('NAME', 'c'),
        ('RIGHT_PARENTHESIS', ')'),
        ('COLON', ':'),
        ('ENDL', '\n', [], [('SPACE', '    ')]),
        ('INDENT', ''),
        ('PASS', 'pass'),
        ('ENDL', '\n'),
        ('DEDENT', ''),
        ('ENDMARKER', ''),
    ], [
        {
            'arguments': [
                {
                    'first_formatting': [],
                    'second_formatting': [],
                    "annotation": {},
                    "annotation_first_formatting": [],
                    "annotation_second_formatting": [],
                    'type': 'def_argument',
                    'target': {
                        'type': 'tuple',
                        'first_formatting': [],
                        'with_parenthesis': True,
                        'fourth_formatting': [],
                        'second_formatting': [],
                        'third_formatting': [],
                        'value': [
                            {
                                'value': 'a',
                                'type': 'name',
                            },
                            {
                                'first_formatting': [],
                                'second_formatting': [],
                                'type': 'comma'
                            },
                            {
                                'type': 'tuple',
                                'first_formatting': [],
                                'with_parenthesis': True,
                                'fourth_formatting': [],
                                'second_formatting': [],
                                'third_formatting': [],
                                'value': [
                                    {
                                        'value': 'x',
                                        'type': 'name',
                                    },
                                    {
                                        'first_formatting': [],
                                        'second_formatting': [],
                                        'type': 'comma'
                                    },
                                    {
                                        'value': 'y',
                                        'type': 'name',
                                    },
                                ]
                            },
                        ]
                    },
                    'value': {
                        'type': 'name',
                        'value': 'c',
                    }
                }
            ],
            'decorators': [],
            'async': False,
            'async_formatting': [],
            "return_annotation": {},
            "return_annotation_first_formatting": [],
            "return_annotation_second_formatting": [],
            'fifth_formatting': [],
            'first_formatting': [
                {
                    'type': 'space',
                    'value': ' '
                }
            ],
            'fourth_formatting': [],
            'name': 'function_name',
            'second_formatting': [],
            'sixth_formatting': [],
            'third_formatting': [],
            'type': 'def',
            'value': [
                {
                    'formatting': [],
                    'indent': '    ',
                    'type': 'endl',
                    'value': '\n'
                },
                {
                    'type': 'pass'
                },
                {
                    'formatting': [],
                    'indent': '',
                    'type': 'endl',
                    'value': '\n'
                }
            ]
        }
    ])


def test_standalone_annotation():
    """
    x :  int
    """
    parse_simple([
        ('NAME', 'x'),
        ('COLON', ':', [('SPACE', ' ')], [('SPACE', '  ')]),
        ('NAME', 'int'),
    ], [
        {
            "type": "standalone_annotation",
            "annotation": {"type": "name", "value": "int"},
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": "  "}],
            "target": {"type": "name", "value": "x"},
        }
    ])


def test_typed_variable_with_assignment():
    """
    x : int = 1
    """
    parse_simple([
        ('NAME', 'x'),
        ('COLON', ':', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('NAME', 'int'),
        ('EQUAL', '=', [('SPACE', ' ')], [('SPACE', ' ')]),
        ('INT', '1'),
    ], [
        {
            "type": "assignment",
            "annotation": {"type": "name", "value": "int"},
            "first_formatting": [{"type": "space", "value": " "}],
            "second_formatting": [{"type": "space", "value": " "}],
            "annotation_first_formatting": [{"type": "space", "value": " "}],
            "annotation_second_formatting": [{"type": "space", "value": " "}],
            "target": {"type": "name", "value": "x"},
            "operator": "",
            "value": {"section": "number", "type": "int", "value": "1"}
        }
    ])



def has_print_function(tokens):
    p = 0
    while p < len(tokens):
        if tokens[p][0] != 'FROM':
            p += 1
            continue
        if tokens[p + 1][0:2] != ('NAME', '__future__'):
            p += 1
            continue
        if tokens[p + 2][0] != 'IMPORT':
            p += 1
            continue

        current = p + 3
        # ignore LEFT_PARENTHESIS token
        if tokens[current][0] == 'LEFT_PARENTHESIS':
            current += 1

        while (current < len(tokens) and tokens[current][0] == 'NAME'):
            if tokens[current][1] == 'print_function':
                return True

            # ignore AS and NAME tokens if present
            # anyway, ignore COMMA token
            if current + 1 < len(tokens) and tokens[current + 1][0] == 'AS':
                current += 4
            else:
                current += 2
        p += 1

    return False


def replace_print_by_name(tokens):
    def is_print(token):
        return token[0] == 'PRINT'

    return [('NAME', 'print') if is_print(x) else x for x in tokens]



from rply.token import BaseBox


class BaronToken(BaseBox):
    """
    Represents a syntactically relevant piece of text.

    :param name: A string describing the kind of text represented.
    :param value: The actual text represented.
    :param source_pos: A :class:`SourcePosition` object representing the
                       position of the first character in the source from which
                       this token was generated.
    """
    def __init__(self, name, value, hidden_tokens_before=None, hidden_tokens_after=None):
        self.name = name
        self.value = value
        self.hidden_tokens_before = list(map(self._translate_tokens_to_ast_node, hidden_tokens_before if hidden_tokens_before else []))
        self.hidden_tokens_after = list(map(self._translate_tokens_to_ast_node, hidden_tokens_after if hidden_tokens_after else []))

    def _translate_tokens_to_ast_node(self, token):
        if token[0] == "ENDL":
            return {
                "type": token[0].lower(),
                "value": token[1],
                "indent": token[3][0][1] if len(token) == 4 and token[3] else "",
                "formatting": list(map(self._translate_tokens_to_ast_node, token[2]) if len(token) >= 3 else []),
            }
        if len(token) >= 3:
            return {
                "type": token[0].lower(),
                "value": token[1],
                "formatting": list(map(self._translate_tokens_to_ast_node, token[2]) if len(token) >= 3 else []),
            }
        if token[0] == "COMMENT":
            return {
                "type": token[0].lower(),
                "value": token[1],
                "formatting": [],
            }
        return {
            "type": token[0].lower(),
            "value": token[1],
        }

    def __repr__(self):
        return "Token(%r, %r, %s, %s)" % (self.name, self.value, self.hidden_tokens_before, self.hidden_tokens_after)

    def __eq__(self, other):
        if not isinstance(other, BaronToken):
            return NotImplemented
        return self.name == other.name and self.value == other.value

    def render(self):
        before = "".join([(x["indent"] if x["type"] == "endl" else "") + x["value"] for x in self.hidden_tokens_before])
        after = "".join([(x["indent"] if x["type"] == "endl" else "") + x["value"] for x in self.hidden_tokens_after])
        # print self.hidden_tokens_before, self.value, self.hidden_tokens_after
        return before + self.value + after

    def gettokentype(self):
        """
        Returns the type or name of the token.
        """
        return self.name

    def getstr(self):
        """
        Returns the string represented by this token.
        """
        return self.value



from .utils import create_node_from_token


def include_primivites(pg, print_function):
    if not print_function:
        @pg.production("print_stmt : PRINT")
        def print_stmt_empty(pack):
            (print_,) = pack
            return {
                "type": "print",
                "value": [],
                "destination": None,
                "destination_formatting": [],
                "formatting": [],
            }

        @pg.production("print_stmt : PRINT testlist")
        def print_stmt(pack):
            (print_, testlist) = pack
            return {
                "type": "print",
                "value": testlist["value"] if testlist["type"] == "tuple" and testlist["with_parenthesis"] is False else [testlist],
                "destination": None,
                "destination_formatting": [],
                "formatting": print_.hidden_tokens_after,
            }

        @pg.production("print_stmt : PRINT RIGHT_SHIFT test")
        def print_stmt_redirect(pack):
            (print_, right_shift, test) = pack
            return {
                "type": "print",
                "value": [],
                "destination": test,
                "destination_formatting": right_shift.hidden_tokens_after,
                "formatting": print_.hidden_tokens_after,
            }

        @pg.production("print_stmt : PRINT RIGHT_SHIFT test COMMA testlist")
        def print_stmt_redirect_testlist(pack):
            (print_, right_shift, test, comma, testlist) = pack
            value = [{
                "type": "comma",
                "first_formatting": comma.hidden_tokens_before,
                "second_formatting": comma.hidden_tokens_after,
            }]
            value += testlist["value"] if testlist["type"] == "tuple" else [testlist]
            return {
                "type": "print",
                "value": value,
                "destination": test,
                "destination_formatting": right_shift.hidden_tokens_after,
                "formatting": print_.hidden_tokens_after,
            }

        @pg.production("assert_stmt : EXEC expr")
        def exec_stmt(pack):
            (exec_, expr) = pack
            return {
                "type": "exec",
                "value": expr,
                "globals": None,
                "locals": None,
                "first_formatting": exec_.hidden_tokens_after,
                "second_formatting": [],
                "third_formatting": [],
                "fourth_formatting": [],
                "fifth_formatting": []
            }

        @pg.production("assert_stmt : EXEC expr IN test")
        def exec_stmt_in(pack):
            (exec_, expr, in_, test) = pack
            return {
                "type": "exec",
                "value": expr,
                "globals": test,
                "locals": None,
                "first_formatting": exec_.hidden_tokens_after,
                "second_formatting": in_.hidden_tokens_before,
                "third_formatting": in_.hidden_tokens_after,
                "fourth_formatting": [],
                "fifth_formatting": []
            }

        @pg.production("assert_stmt : EXEC expr IN test COMMA test")
        def exec_stmt_in_comma(pack):
            (exec_, expr, in_, test, comma, test2) = pack
            return {
                "type": "exec",
                "value": expr,
                "globals": test,
                "locals": test2,
                "first_formatting": exec_.hidden_tokens_after,
                "second_formatting": in_.hidden_tokens_before,
                "third_formatting": in_.hidden_tokens_after,
                "fourth_formatting": comma.hidden_tokens_before,
                "fifth_formatting": comma.hidden_tokens_after
            }

    @pg.production("flow_stmt : return_stmt")
    @pg.production("flow_stmt : break_stmt")
    @pg.production("flow_stmt : continue_stmt")
    @pg.production("flow_stmt : yield_stmt")
    @pg.production("yield_stmt : yield_expr")
    def flow(pack):
        (flow_stmt,) = pack
        return flow_stmt

    @pg.production("return_stmt : RETURN")
    def return_empty(pack):
        (token,) = pack
        return {
            "type": token.name.lower(),
            "value": None,
            "formatting": token.hidden_tokens_after,
        }

    @pg.production("yield_expr : YIELD")
    def yield_expr(pack):
        (yield_,) = pack
        return {
            "type": yield_.name.lower(),
            "value": None,
            "formatting": yield_.hidden_tokens_after,
        }

    @pg.production("break_stmt : BREAK")
    @pg.production("continue_stmt : CONTINUE")
    @pg.production("pass_stmt : PASS")
    def break_stmt(pack):
        (token,) = pack
        return {"type": token.name.lower()}

    @pg.production("raise_stmt : RAISE")
    def raise_stmt_empty(pack):
        (raise_,) = pack
        return {
            "type": "raise",
            "value": None,
            "instance": None,
            "traceback": None,
            "first_formatting": raise_.hidden_tokens_after,
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "comma_or_from": None,
        }

    @pg.production("raise_stmt : RAISE test")
    def raise_stmt(pack):
        (raise_, test) = pack
        return {
            "type": "raise",
            "value": test,
            "instance": None,
            "traceback": None,
            "first_formatting": raise_.hidden_tokens_after,
            "second_formatting": [],
            "third_formatting": [],
            "fourth_formatting": [],
            "fifth_formatting": [],
            "comma_or_from": None,
        }

    @pg.production("raise_stmt : RAISE test FROM test")
    def raise_stmt_from(pack):
        (raise_, test, from_, test2) = pack
        return {
            "type": "raise",
            "value": test,
            "instance": test2,
            "traceback": None,
            "first_formatting": raise_.hidden_tokens_after,
            "second_formatting": from_.hidden_tokens_before,
            "third_formatting": from_.hidden_tokens_after,
            "fourth_formatting": [],
            "fifth_formatting": [],
            "comma_or_from": "from",
        }

    @pg.production("raise_stmt : RAISE test COMMA test")
    def raise_stmt_instance(pack):
        (raise_, test, comma, test2) = pack
        return {
            "type": "raise",
            "value": test,
            "instance": test2,
            "traceback": None,
            "first_formatting": raise_.hidden_tokens_after,
            "second_formatting": comma.hidden_tokens_before,
            "third_formatting": comma.hidden_tokens_after,
            "fourth_formatting": [],
            "fifth_formatting": [],
            "comma_or_from": ",",
        }

    @pg.production("raise_stmt : RAISE test COMMA test COMMA test")
    def raise_stmt_instance_traceback(pack):
        (raise_, test, comma, test2, comma2, test3) = pack
        return {
            "type": "raise",
            "value": test,
            "instance": test2,
            "traceback": test3,
            "first_formatting": raise_.hidden_tokens_after,
            "second_formatting": comma.hidden_tokens_before,
            "third_formatting": comma.hidden_tokens_after,
            "fourth_formatting": comma2.hidden_tokens_before,
            "fifth_formatting": comma2.hidden_tokens_after,
            "comma_or_from": ",",
        }

    @pg.production("assert_stmt : ASSERT test")
    def assert_stmt(pack):
        (assert_, test) = pack
        return {
            "type": "assert",
            "value": test,
            "message": None,
            "first_formatting": assert_.hidden_tokens_after,
            "second_formatting": [],
            "third_formatting": []
        }

    @pg.production("assert_stmt : ASSERT test COMMA test")
    def assert_stmt_message(pack):
        (assert_, test, comma, test2) = pack
        return {
            "type": "assert",
            "value": test,
            "message": test2,
            "first_formatting": assert_.hidden_tokens_after,
            "second_formatting": comma.hidden_tokens_before,
            "third_formatting": comma.hidden_tokens_after
        }

    @pg.production("global_stmt : GLOBAL names")
    def global_stmt(pack):
        (global_, names) = pack
        return {
            "type": "global",
            "formatting": global_.hidden_tokens_after,
            "value": names,
        }

    @pg.production("nonlocal_stmt : NONLOCAL names")
    def nonlocal_stmt(pack):
        (token, names) = pack
        return {
            "type": "nonlocal",
            "formatting": token.hidden_tokens_after,
            "value": names,
        }

    @pg.production("names : NAME")
    def names_name(pack):
        (name,) = pack
        return [create_node_from_token(name)]

    @pg.production("names : names comma name")
    def names_names_name(pack):
        (names, comma, name,) = pack
        return names + [comma, name]

    @pg.production("return_stmt : RETURN testlist")
    @pg.production("yield_expr : YIELD testlist")
    @pg.production("del_stmt : DEL exprlist")
    def return_testlist(pack):
        (token, testlist) = pack
        return {
            "type": token.name.lower(),
            "value": testlist,
            "formatting": token.hidden_tokens_after,
        }

    @pg.production("yield_expr : YIELD FROM test")
    def yield_from_expr(pack):
        (yield_, from_, test) = pack
        return {
            "type": "yield_from",
            "first_formatting": from_.hidden_tokens_after,
            "value": test,
            "formatting": yield_.hidden_tokens_after,
        }

    @pg.production("lambdef : LAMBDA COLON test")
    @pg.production("old_lambdef : LAMBDA COLON old_test")
    def lambdef(pack):
        (lambda_, colon, test) = pack
        return {
            "type": "lambda",
            "arguments": [],
            "first_formatting": lambda_.hidden_tokens_after,
            "second_formatting": colon.hidden_tokens_before,
            "third_formatting": colon.hidden_tokens_after,
            "value": test
        }

    @pg.production("lambdef : LAMBDA parameters COLON test")
    @pg.production("old_lambdef : LAMBDA parameters COLON old_test")
    def lambdef_arguments(pack):
        (lambda_, parameters, colon, test) = pack
        return {
            "type": "lambda",
            "arguments": parameters,
            "first_formatting": lambda_.hidden_tokens_after,
            "second_formatting": colon.hidden_tokens_before,
            "third_formatting": colon.hidden_tokens_after,
            "value": test
        }



from baron import parse, dumps, tokenize


def test_regression_trailing_comment_after_colon():
    assert parse("def a(): # pouf\n    pass")


def test_regression_trailing_comment_after_colon_no_space():
    assert parse("def a():# pouf\n    pass")


def test_regression_trailing_comment_after_colon_dump():
    code = "def a(): # pouf\n    pass\n"
    assert dumps(parse(code)) == code


def test_regression_trailing_comment_after_colon_no_space_dump():
    code = "def a():# pouf\n    pass\n"
    assert dumps(parse(code)) == code


def test_comment_in_middle_of_ifelseblock():
    code = 'if a:\n    pass\n# comment\nelse:\n    pass\n'
    assert dumps(parse(code)) == code


def test_new_float_notation():
    code = 'a._'
    assert tokenize(code)[:-1] == [('NAME', 'a'), ('DOT', '.'), ('NAME', '_')]



from baron.future import has_print_function, replace_print_by_name
from baron.baron import parse, tokenize
import json


def print_token_is_a_function(code):
    return has_print_function(tokenize(code))


def print_is_parsed_as_a_function(parsed_code):
    code_json = json.dumps(parsed_code)
    return '"type": "name", "value": "print"' in code_json \
           or '"value": "print", "type": "name"' in code_json


def test_no_future():
    code = "a = 1"
    assert not print_token_is_a_function(code)


def test_other_future():
    code = "from __future__ import other"
    assert not print_token_is_a_function(code)


def test_print_future():
    code = "from __future__ import print_function"
    assert print_token_is_a_function(code)


def test_print_future_as():
    code = "from __future__ import print_function as p_f"
    assert print_token_is_a_function(code)


def test_print_future_comma():
    code = "from __future__ import a, b, print_function"
    assert print_token_is_a_function(code)


def test_print_future_comma_as():
    code = "from __future__ import a as c, b as d, print_function as e"
    assert print_token_is_a_function(code)


def test_print_no_future_comma_as():
    code = "from __future__ import a as c, b as d"
    assert not print_token_is_a_function(code)


def test_print_future_in_parenthesis():
    code = "from __future__ import (a, b, print_function)"
    assert print_token_is_a_function(code)


def test_print_future_in_parenthesis_as():
    code = "from __future__ import (a as c, b as d, print_function as e)"
    assert print_token_is_a_function(code)


def test_print_no_future_in_parenthesis_as():
    code = "from __future__ import (a as c, b as d)"
    assert not print_token_is_a_function(code)


def test_print_future_second():
    code = """from __future__ import a, b as e
from __future__ import c, print_function"""
    assert print_token_is_a_function(code)


def test_auto_print_as_name():
    code = "from __future__ import print_function\nprint(a)"
    assert print_is_parsed_as_a_function(parse(code))


def test_auto_print_as_print():
    code = "print(a)"
    assert not print_is_parsed_as_a_function(parse(code))


def test_print_as_name():
    code = "print(a)"
    assert print_is_parsed_as_a_function(parse(code, True))


def test_print_as_print():
    code = "print(a)"
    assert not print_is_parsed_as_a_function(parse(code, False))


def test_replace_print_token():
    tokens = [('PRINT', 'print'), ('LEFT_PARENTHESIS', '('), ('NAME', 'A'), ('RIGHT_PARENTHESIS', ')'), ('ENDMARKER', '')]
    after = [('NAME', 'print'), ('LEFT_PARENTHESIS', '('), ('NAME', 'A'), ('RIGHT_PARENTHESIS', ')'), ('ENDMARKER', '')]
    assert after == replace_print_by_name(tokens)
